# -*- coding: utf-8 -*-
import glob
import law
import luigi
import logging
import os
import re
import shutil
import six

from functools import partial
from fnmatch import fnmatch
from subprocess import PIPE

from law.util import colored

from analysis.tasks.Warmup import Warmup
from analysis.tasks.MergeFastWarm import MergeFastWarm

from analysis.framework import Task, HTCondorWorkflow, SpaceSeparatedListParameter
from analysis.mixins import TarballExtractionMixin
from analysis.util import createRuncard, interruptable_popen, working_directory

logger = logging.getLogger(__name__)
working_directory = partial(working_directory, logger=logger)


class FastProd(Task, TarballExtractionMixin, HTCondorWorkflow, law.LocalWorkflow):
    base_runcard = luigi.Parameter()
    steer_dir = luigi.Parameter()

    default_channels = ["LO", "R", "V", "RRa", "RRb", "RV", "VV"]
    channels = SpaceSeparatedListParameter(
        default=default_channels,
        cls=luigi.Parameter,
        choices=("LO", "R", "V", "RRa", "RRb", "RV", "VV")
    )
    fastprod_events = SpaceSeparatedListParameter(cls=luigi.IntParameter)
    fastprod_jobs = SpaceSeparatedListParameter(cls=luigi.IntParameter)
    starting_seeds = SpaceSeparatedListParameter(cls=luigi.IntParameter)

    def create_branch_map(self):
        """Branch data dicts. Each branch should correspond to exactly one output."""
        i = 0
        branchmap = {}
        for j, channel in enumerate(self.channels):
            for k in six.moves.range(int(self.fastprod_jobs[j])):
                branchmap[i] = {
                    'index': j,
                    'channel': channel,
                    'events': self.fastprod_events[j],
                    'seed': str(k + int(self.starting_seeds[j]))
                }
                i += 1
        return branchmap

    # TODO: Remove warmup dependency?
    # It is preferable to have production jobs fail for missing warmups than
    # having such warmups reproduced on-the-fly for either a long time or
    # with very low precision.
    def workflow_requires(self):
        return {'warmup': Warmup(), 'fastwarm': MergeFastWarm()}

    def requires(self):
        return {'warmup': Warmup(branch=self.branch_data['index']), 'fastwarm': MergeFastWarm()}

    def output(self):
        return self.remote_or_local_target(
            '{}.{}.{}.fastprod.s{}.tar.gz'.format(
                self.process,
                self.branch_data['channel'],
                self.name,
                self.branch_data['seed'],
            ),
        )

    def run(self):
        rundirpath = os.getcwd()
        tmpdirpath = os.path.join(rundirpath, 'tmpdir{}'.format(self.branch_data['seed']))

        logger.debug("Running job in directory: %s", rundirpath)
        outfile = os.path.basename(self.output().path)
        logfile = re.sub('.tar.gz$', '.log', outfile)
        runfile = re.sub('.tar.gz$', '.run', outfile)

        job_ok = True  # set to False if exception is raised

        try:
            # -- create temporary directory
            if not os.path.exists(tmpdirpath):
                logger.debug("Creating temporary directory: %s", tmpdirpath)
                os.mkdir(tmpdirpath)
            elif os.path.isdir(tmpdirpath):
                logger.warning("Temporary directory already exists: %s", tmpdirpath)
            else:
                raise Exception("Could not create temporary directory: %s".format(tmpdirpath))

            # -- check and copy base run card
            logger.debug("Checking for base runcard: %s", self.base_runcard)
            if not os.path.isfile(self.base_runcard):
                raise Exception('Base runcard {} not found!'.format(self.base_runcard))
            logger.debug(
                "Copying base runcard: %s -> %s", self.base_runcard,
                os.path.join(tmpdirpath, self.base_runcard)
            )
            shutil.copy(self.base_runcard, os.path.join(tmpdirpath, self.base_runcard))

            # -- check and copy steering file
            logger.debug("Checking for steer directory: %s", self.steer_dir)
            if not os.path.isdir(self.steer_dir):
                raise Exception('Steerfile directory {} not found!'.format(self.steer_dir))

            with working_directory(self.steer_dir):
                strfiles = glob.glob('*.str')
                logger.debug("Number of steering files found: %d", len(strfiles))
                if not len(strfiles) > 0:
                    raise Exception('Steerfile directory {} empty, aborted.'.format(self.steer_dir))

                for fname in strfiles:
                    logger.debug(
                        "Copying steering file: %s -> %s", fname,
                        os.path.join(rundirpath, tmpdirpath, fname)
                    )
                    shutil.copy(fname, os.path.join(rundirpath, tmpdirpath, fname))

            # ensure output parent directory exists
            logger.debug("Creating output parent directory: %s", self.output().parent.path)
            self.output().parent.touch()

            # run all subsequent command in tmpdir
            with working_directory(tmpdirpath):
                # extract NNLOJET warmup result tarball to tmpdir
                logger.debug("Loading NNLOJET warmup file from: %s", self.input()['warmup'].path)
                self.extract_tarball(
                    self.input()['warmup'],
                    law.LocalDirectoryTarget(tmpdirpath),
                    formatter="tar",
                    logger=logger,
                )

                # extract fastNLO warmup result tarball to tmpdir
                logger.debug("Loading fastNLO warmup file from: %s", self.input()['fastwarm'].path)
                self.extract_tarball(
                    self.input()['fastwarm'],
                    law.LocalDirectoryTarget(tmpdirpath),
                    formatter="tar",
                    logger=logger,
                )

                logger.debug("Out filename is: %s", outfile)
                logger.debug("Log filename is: %s", logfile)
                logger.debug("Run cardname is: %s", runfile)

                logger.debug("Perfoming in-place @@ substitution on runcard: %s", runfile)
                with open(runfile, 'w') as f:
                    runcard = createRuncard(
                        open(self.base_runcard, 'r').read(), {
                            'channel': self.branch_data['channel'],
                            'events': self.branch_data['events'],
                            'seed': self.branch_data['seed'],
                            'iterations': '1',
                            'warmup': '0',
                            'production': '.true.',
                            'unit_phase': ''
                        }
                    )
                    f.write(runcard)

                logger.debug("Starting NNLOJET with runcard: %s", runfile)

                command_line_args = ['NNLOJET', '-run', runfile]
                logger.info("Running command: %s",
                    colored(' '.join(command_line_args), 'yellow'))
                code, out, error = interruptable_popen(command_line_args,
                                                       stdout=PIPE,
                                                       stderr=PIPE)
                logger.debug("Return code: %d", code)
                logger.debug("Standard output:\n%s\n", out)
                logger.debug("Standard error:\n%s\n", error)

                logger.debug("Dumping standard out to file:  %s", logfile)
                with open(logfile, 'w') as f:
                    f.write(out)

                if (code == 0):
                    logger.info("NNLOJET status: SUCCESS")
                    job_ok = True
                else:
                    raise RuntimeError("NNLOJET status: FAIL (exit code {})".format(code))

        # mark job as failed on error, make log entry and reraise
        except Exception as e:
            job_ok = False
            logger.error('%s raised during job execution: %s', e.__class__.__name__, e)
            raise

        # finalize job
        finally:
            logger.debug("Finalizing job...")

            with working_directory(tmpdirpath):
                logger.debug("Files in current directory:\n  %s", '\n  '.join(glob.glob('*')))

                # write out results on success, debug info on fail
                if job_ok:
                    tarfilter = lambda n: n if (
                        n.name == '.' or
                        fnmatch(n.name, '*.wrm') or fnmatch(os.path.basename(n.name), logfile) or
                        fnmatch(os.path.basename(n.name), runfile) or fnmatch(n.name, '*.str') or
                        fnmatch(n.name, '*.wrm') or fnmatch(n.name, '*.vBa') or
                        fnmatch(n.name, '*.vRa') or fnmatch(n.name, '*.RRa') or
                        fnmatch(n.name, '*.RRb') or fnmatch(n.name, '*.tab.gz') or
                        fnmatch(n.name, '*.dat')
                    ) else None

                    logger.debug(
                        "Job successful. Writing output tarfile from directory: %s", os.getcwd()
                    )
                    self.output().dump(os.getcwd(), formatter="tar", add_kwargs=dict(filter=tarfilter))
                else:
                    debugfile = re.sub(
                        '.tar.gz$', '.debug.tar.gz', os.path.basename(self.output().path)
                    )
                    logger.debug(
                        "Job failed. Writing debug tarfile: %s -> %s",
                        os.getcwd(), debugfile
                    )
                    self.remote_target(debugfile).dump(os.getcwd(), formatter="tar")

            logger.debug("Removing temporary directory: %s", tmpdirpath)
            shutil.rmtree(tmpdirpath)
