# -*- coding: utf-8 -*-
import glob
import law
import luigi
import logging
import os

from itertools import product
from subprocess import PIPE

from law.util import colored

from analysis.tasks.CopyTables import CopyTables

from analysis.framework import Task, SpaceSeparatedListParameter
from analysis.util import interruptable_popen

logger = logging.getLogger(__name__)


class PlotRuntime(Task, law.LocalWorkflow):
    command = 'fastnnlo_runtime.py'

    force = luigi.BoolParameter(default=False)

    merge_dir = luigi.Parameter(default='grids')
    plots_dir = luigi.Parameter(default='plots')

    default_channels = ["LO", "R", "V", "RRa", "RRb", "RV", "VV"]
    channels = SpaceSeparatedListParameter(
        default=default_channels,
        cls=luigi.Parameter,
        choices=("LO", "R", "V", "RRa", "RRb", "RV", "VV")
    )
    jobnameext = luigi.Parameter()

    def create_branch_map(self):
        """Branch data dicts. Each branch should correspond to exactly one output."""
        return {
            i : {'channel': ch, 'glob': "*" if ch == 'ALL' else ch}
            for i, ch in enumerate(self.channels + ('ALL',))
        }

    def workflow_requires(self):
        if self.force:
            return {"copytables": CopyTables.req(self, acceptance=0)}
        else:
            return {"copytables": CopyTables.req(self)}

    def output(self):
        out_dir = os.getenv("ANALYSIS_PATH") or os.getcwd()
        return law.LocalFileTarget(
            '{}/{}/{}/{}/{}.{}-{}.runtime.png'.format(
                out_dir,
                self.plots_dir, self.name, self.__class__.__name__,
                self.process, self.branch_data['channel'], self.jobnameext
            )
        )

    def run(self):
        # ensure output parent directory exists
        logger.debug("Creating output parent directory: %s", self.output().parent.path)
        self.output().parent.touch()

        outpath = self.output().parent.path
        logger.debug("Output path: %s", outpath)

        plotnam = "{}.{}-{}".format(
            self.process, self.branch_data['channel'], self.jobnameext
        )
        logglob = os.path.join(
            self.merge_dir, self.name, self.branch_data['glob'],
            "{}.{}.*.fastprod.s*.log".format(self.process, self.branch_data['glob'])
        )
        logger.debug("Log glob: %s", logglob)

        command_line_args = [
            self.command,
            str(logglob),
            '-o', outpath,
            '-f', plotnam,
            '--CPUtime'
        ]
        if self.branch_data['channel'] == 'ALL':
            command_line_args.append('--Events')

        job_ok = True

        try:
            logger.info("Running command: %s",
                    colored(' '.join(command_line_args), 'yellow'))
            code, out, error = interruptable_popen(command_line_args,
                                                   stdout=PIPE,
                                                   stderr=PIPE)

            logger.debug("Return code: %d", code)
            logger.debug("Standard output:\n%s\n", out)
            logger.debug("Standard error:\n%s\n", error)

            if (code == 0):
                logger.info("%s status: SUCCESS", self.command)
                logger.info("fastnnlo_runtime.py status: SUCCESS")
            else:
                raise RuntimeError("{} status: FAIL (exit code {})".format(self.command, code))

        # mark job as failed on error, make log entry and reraise
        except Exception as e:
            job_ok = False
            logger.error('%s raised during job execution: %s', e.__class__.__name__, e)
            raise

        finally:
            if not job_ok and self.output().exists():
                logger.info("Removing failed task output: %s", self.output().path)
                self.output().remove()
